import numpy as np
from cuqi.geometry import _get_identity_geometries
from cuqi.distribution import Distribution
from cuqi.distribution import Gaussian
import warnings

class Lognormal(Distribution):
    """
    Multivariate Lognormal distribution

    Parameters
    ------------
    mean: np.ndarray
        Mean of the normal distribution used to define the lognormal distribution 

    cov: np.ndarray
        Covariance matrix of the normal distribution used to define the lognormal distribution 
    
    Example
    -------
    .. code-block:: python
    
        # Generate a lognormal distribution
        mean = np.array([1.5,1])
        cov = np.array([[3, 0],[0, 1]])
        x = cuqi.distribution.Lognormal(mean, cov)
        samples = x.sample(10000)
        samples.hist_chain(1, bins=70)

    """
    def __init__(self, mean, cov, is_symmetric=False, **kwargs):
        super().__init__(is_symmetric=is_symmetric, **kwargs) 
        self.mean = mean
        self.cov = cov
        self._normal = Gaussian(self.mean, self.cov)

    @property
    def _normal(self):
        if not np.all(self._Gaussian.mean == self.mean):
            self._Gaussian.mean = self.mean
        if not np.all(self._Gaussian.cov == self.cov):
            self._Gaussian.cov = self.cov 
        return self._Gaussian

    @_normal.setter
    def _normal(self, value):
        self._Gaussian = value

    @property
    def dim(self):
        return self._normal.dim

    def pdf(self, x):
        if np.any(x<=0):
            return 0
        else:
            return self._normal.pdf(np.log(x))*np.prod(1/x)

    def logpdf(self, x):
        return np.log(self.pdf(x))

    def _gradient(self, val, *args, **kwargs):
        #Avoid complicated geometries that change the gradient.
        if not type(self.geometry) in _get_identity_geometries():
            raise NotImplementedError("Gradient not implemented for distribution {} "
                                      "with geometry {}".format(self,self.geometry))

        elif not callable(self._normal.mean): # for prior
            return np.diag(1/val)@(-1+self._normal.gradient(np.log(val)))
        elif hasattr(self.mean,"gradient"): # for likelihood
            model = self._normal.mean
            dev = np.log(val) - model.forward(*args, **kwargs)
            return  model.gradient(self._normal.prec@dev, *args, **kwargs) # Jac(x).T@(self._normal.prec@dev)
        else:
            warnings.warn('Gradient not implemented for {}'.format(type(self._normal.mean)))

    def _sample(self, N=1, rng=None):
        return np.exp(self._normal._sample(N,rng))



import cuqi
import numpy as np

import pytest
from pytest import approx
from scipy.ndimage import convolve

@pytest.mark.parametrize("dim,kernel,kernel_param,expected",[
    (128,"Gauss",None,45.3148),
    (36,"Gauss",None,12.7448),
    (36,"Gauss",5,18.0238),
    (36,"sinc",None,9.2320),
    (36,"sinc",37,6.0585),
    (36,"prolate",None,9.2320),
    (36,"prolate",37,6.0585),
    (36,"vonmises",None,12.8714),
    (36,"vonmises",37,7.7606),
])
def test_Deconvolution_MatrixNorm_regression(dim,kernel,kernel_param,expected):
    tp = cuqi.testproblem.Deconvolution1D(dim=dim, PSF=kernel, PSF_param=kernel_param, use_legacy=True)
    assert np.linalg.norm(tp.model.get_matrix()) == approx(expected,rel=1e-4)

@pytest.mark.parametrize("dim,phantom,phantom_param,expected",[
    (36,"Gauss",None,2.0944),
    (36,"Gauss",3,2.7039),
    (128,"Gauss",10,2.8211),
    (36,"sinc",None,1.8518),
    (36,"vonmises",None,2.1582),
    (36,"square",None,2),
    (36,"hat",None,1.2247),
    (36,"bumps",None,2.4009),
    (36,"derivgauss",None,2.4858),                            
])
def test_Deconvolution_PhantomNorm_regression(dim,phantom,phantom_param,expected):
    tp = cuqi.testproblem.Deconvolution1D(dim=dim,phantom=phantom,phantom_param=phantom_param)
    assert np.linalg.norm(tp.exactSolution) == approx(expected,rel=1e-4)

@pytest.mark.parametrize("dim,phantom,phantom_param,expected",[
    (35,"gauss",None,2.0643),   
    (35,"sinc",None,1.8252),
    (35,"vonmises",None,2.0848),
    (35,"square",None,2),
    (35,"hat",None,1.2247),
    (35,"bumps",None,2.3673),
    (35,"derivgauss",None,2.5478),  
])
def test_Deconvolution_PhantomNorm2_regression(dim,phantom,phantom_param,expected):
    x = cuqi.testproblem._getExactSolution(dim=dim,phantom=phantom,phantom_param=phantom_param)
    assert np.linalg.norm(x) == approx(expected,rel=1e-4)

@pytest.mark.parametrize("noise_type,noise_std,expected",[
    ("scaledgaussian",0.2,74.5626529810876),   
    ("scaledgaussian",3,235.68073267350434),   
    ("gaussian",0.2,75.7589097857315),  
    ("gaussian",3,79.76620602993175),  
])
def test_Deconvolution_noise_regression(noise_type,noise_std,expected):
    np.random.seed(1337)
    tp = cuqi.testproblem.Deconvolution1D(noise_type=noise_type,noise_std=noise_std, use_legacy=True, phantom="gauss")
    assert np.linalg.norm(tp.data) == approx(expected)

def test_Deconvolution_custom_kernel():
    """ Tests using a custom kernel. """
    grid = np.linspace(-1, 1, 128)
    kernel = np.exp(-(15*grid)**2)

    TP = cuqi.testproblem.Deconvolution1D(dim=128, PSF=kernel, phantom="pc", use_legacy=True, noise_std=0.05)

    # Test expected value
    assert np.linalg.norm(TP.exactData) == pytest.approx(83.40353338231859)

    # Test raising of error if kernel is not 1D
    with pytest.raises(ValueError, match=r"PSF must be a 1D array"):
        TP = cuqi.testproblem.Deconvolution1D(dim=128, PSF=np.ones((128, 128)))


def test_Deconvolution_custom_phantom():
    """ Tests using a custom phantom. """
    grid = np.linspace(-1, 1, 128)
    phantom = np.sin(grid)

    TP = cuqi.testproblem.Deconvolution1D(dim=128, phantom=phantom)

    # Test expected value
    assert np.linalg.norm(TP.exactSolution) == approx(5.94476801594791)

    # Test raising of error if phantom is not 1D
    with pytest.raises(ValueError, match=r"phantom must be a 1D array"):
        TP = cuqi.testproblem.Deconvolution1D(dim=128, phantom=np.ones((128, 128)))

    # Test raising of error if phantom is not same length as dim
    with pytest.raises(ValueError, match=r"phantom must be a 1D array of length dim"):
        TP = cuqi.testproblem.Deconvolution1D(dim=128, phantom=np.ones(125))


@pytest.mark.parametrize("model_class", [cuqi.testproblem.Poisson1D, cuqi.testproblem.Heat1D])
@pytest.mark.parametrize("observation_map, expected_grid_equal",
                         [(None, True),
                          (lambda x: x[np.where(x > np.pi/2)], False)])
def test_testproblem_pde_grid_obs(model_class, observation_map, expected_grid_equal):
    """ Test that the observation operator works on grids or not. """
    N = 128
    L = np.pi

    model = model_class(dim=N, endpoint=L, field_type="Step",
                        observation_grid_map=observation_map).model

    assert model.pde.grids_equal == expected_grid_equal

def test_Poisson():
    # %% Poisson
    N = 129 # spatial discretization 
    L = np.pi   # Length of domain
    amp_fact = 10
    f = lambda xs: 10*np.exp( -( (xs - 0.5)**2 ) / 0.02)
    map = lambda x: np.exp(amp_fact*x)
    dx = L/(N-1)
    x = np.linspace(dx/2,L-dx/2,N)
    true_kappa = np.exp( 5*x*np.exp(-2*x)*np.sin(L-x) )
    model = cuqi.testproblem.Poisson1D(dim=N, endpoint=L, source=f, field_type="KL", map=map).model

    assert np.linalg.norm(model.forward(true_kappa, is_par=False)) == approx(6.183419642601269)
    assert np.linalg.norm(model.forward(np.ones(model.domain_dim))) == approx(5.195849938761418)

def test_Heat():
    # %% HEAT
    N = 128
    L = np.pi
    dx = L/(N+1)
    x = np.linspace(dx,L,N,endpoint=False)
    true_kappa = x*np.exp(-2*x)*np.sin(L-x)
    model = cuqi.testproblem.Heat1D(dim=N, endpoint=L, field_type="KL").model

    assert np.linalg.norm(model.forward(true_kappa, is_par=False)) == approx(0.5476375563249378)
    assert np.linalg.norm(model.forward(np.ones(model.domain_dim))) == approx(0.548657107830281)

def test_Abel():
    N = 128
    L = 1
    h = L/N
    tvec = np.linspace(h/2, L-h/2, N)
    true_image = np.sin(tvec*np.pi)*np.exp(-2*tvec)
    KL_map = lambda x: 10*x
    model = cuqi.testproblem.Abel1D(dim=N, endpoint=L, field_type="KL", KL_map=KL_map).model

    assert np.linalg.norm(model.forward(true_image, is_par=False)) == approx(4.580752014966276)
    assert np.linalg.norm(model.forward(np.ones(model.domain_dim))) == approx(9.37456136258068)

#Deconv 2D tests
#TODO. Add tests for custom PSF
@pytest.mark.parametrize("prior",[
    (cuqi.distribution.Gaussian(np.zeros(128**2), 1, name="x")),
    #(cuqi.distribution.LMRF(0, 1, "zeros", geometry=128**2)),
    #(cuqi.distribution.CMRF(np.zeros(128**2), 1, "zeros")),
])
def test_Deconvolution2D_Sampling_prior(prior): 
    tp = cuqi.testproblem.Deconvolution2D(prior=prior)
    tp.prior.geometry = tp.model.domain_geometry
    tp.sample_posterior(10) # Tests that sampling at least runs withour error.
    #TODO. Make into regression tests + other samplers. Move some tests to BayesianProblem

@pytest.mark.parametrize("phantom",[
    ( np.zeros((128,128)) ),    #Correct size
    ( np.zeros(128**2) ),       #Correct size, but vector
    ( np.zeros((64,64)) ),      #Different size. Should get reshaped
    ( np.zeros(64**2) ),        #Different size. Should get reshaped
    ( "satellite" ),            #String for specific cases
])
def test_Deconvolution2D_phantom(phantom):
    TP= cuqi.testproblem.Deconvolution2D(dim=128, phantom=phantom)
    assert TP.exactSolution.shape == (128**2,)

@pytest.mark.parametrize("PSF_size",[20, 21])
@pytest.mark.parametrize("PSF",["Gauss", "Moffat", "Defocus"])
@pytest.mark.parametrize("BC_MAP",[ 
    ["periodic", "wrap"],
    ["zero", "constant"],
    ["mirror", "mirror"],
    ["Neumann", "reflect"],
    ["nearest", "nearest"],
    ])
def test_Deconvolution2D_convolve(BC_MAP, PSF, PSF_size):
    """ Compare Convolve with ndimage slow convolve"""
    # Define testproblem
    TP = cuqi.testproblem.Deconvolution2D(BC=BC_MAP[0], PSF=PSF, PSF_size=PSF_size)

    # Extract kernel + image + data
    P = TP.Miscellaneous["PSF"]
    x = TP.exactSolution.funvals # Image
    y = TP.exactData.funvals # Data

    # ndimage convolve forward and adjoint (non-FFT)
    forward = lambda x: convolve(x, P, mode=BC_MAP[1])
    adjoint = lambda y: convolve(y, np.flipud(np.fliplr(P)), mode=BC_MAP[1])

    # Compare forward
    assert np.allclose(forward(x), TP.model.forward(x).funvals)
    
    # Compare adjoint
    assert np.allclose(adjoint(y), TP.model.adjoint(y).funvals)

def test_WangCubic():
    """ Test the WangCubic testproblem with simple regression test"""
    TP = cuqi.testproblem.WangCubic()
    assert TP.posterior.logd([0.7, 1.3]) == approx(-119.4710156)
    assert np.allclose(TP.posterior.gradient([0.7, 1.3]), np.array([[26.174, -153.5]]))

    


import numpy as np
import numpy as np
from cuqi.experimental.mcmc import Sampler
from cuqi.array import CUQIarray
from numbers import Number

class NUTS(Sampler):
    """No-U-Turn Sampler (Hoffman and Gelman, 2014).

    Samples a distribution given its logpdf and gradient using a Hamiltonian
    Monte Carlo (HMC) algorithm with automatic parameter tuning.

    For more details see: See Hoffman, M. D., & Gelman, A. (2014). The no-U-turn
    sampler: Adaptively setting path lengths in Hamiltonian Monte Carlo. Journal
    of Machine Learning Research, 15, 1593-1623.

    Parameters
    ----------
    target : `cuqi.distribution.Distribution`
        The target distribution to sample. Must have logpdf and gradient method.
        Custom logpdfs and gradients are supported by using a
        :class:`cuqi.distribution.UserDefinedDistribution`.
    
    initial_point : ndarray
        Initial parameters. *Optional*. If not provided, the initial point is 
        an array of ones.

    max_depth : int
        Maximum depth of the tree >=0 and the default is 15.

    step_size : None or float
        If step_size is provided (as positive float), it will be used as initial
        step size. If None, the step size will be estimated by the sampler.

    opt_acc_rate : float
        The optimal acceptance rate to reach if using adaptive step size.
        Suggested values are 0.6 (default) or 0.8 (as in stan). In principle,
        opt_acc_rate should be in (0, 1), however, choosing a value that is very
        close to 1 or 0 might lead to poor performance of the sampler.

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.

    Example
    -------
    .. code-block:: python

        # Import cuqi
        import cuqi

        # Define a target distribution
        tp = cuqi.testproblem.WangCubic()
        target = tp.posterior

        # Set up sampler
        sampler = cuqi.experimental.mcmc.NUTS(target)

        # Sample
        sampler.warmup(5000)
        sampler.sample(10000)

        # Get samples
        samples = sampler.get_samples()

        # Plot samples
        samples.plot_pair()

    After running the NUTS sampler, run diagnostics can be accessed via the 
    following attributes:

    .. code-block:: python

        # Number of tree nodes created each NUTS iteration
        sampler.num_tree_node_list

        # Step size used in each NUTS iteration
        sampler.epsilon_list

        # Suggested step size during adaptation (the value of this step size is
        # only used after adaptation).
        sampler.epsilon_bar_list

    """

    _STATE_KEYS = Sampler._STATE_KEYS.union({'_epsilon', '_epsilon_bar',
                                                '_H_bar',
                                                'current_target_logd',
                                                'current_target_grad',
                                                'max_depth'})

    _HISTORY_KEYS = Sampler._HISTORY_KEYS.union({'num_tree_node_list',
                                                    'epsilon_list',
                                                    'epsilon_bar_list'})

    def __init__(self, target=None, initial_point=None, max_depth=None,
                 step_size=None, opt_acc_rate=0.6, **kwargs):
        super().__init__(target, initial_point=initial_point, **kwargs)

        # Assign parameters as attributes
        self.max_depth = max_depth
        self.step_size = step_size
        self.opt_acc_rate = opt_acc_rate


    def _initialize(self):

        self._current_alpha_ratio = np.nan # Current alpha ratio will be set to some
                                           # value (other than np.nan) before 
                                           # being used

        self.current_target_logd, self.current_target_grad = self._nuts_target(self.current_point)

        # Parameters dual averaging
        # Initialize epsilon and epsilon_bar
        # epsilon is the step size used in the current iteration
        # after warm up and one sampling step, epsilon is updated
        # to epsilon_bar for the remaining sampling steps.
        if self.step_size is None:
            self._epsilon = self._FindGoodEpsilon()
            self.step_size = self._epsilon
        else:
            self._epsilon = self.step_size

        self._epsilon_bar = "unset"

        # Parameter mu, does not change during the run
        self._mu = np.log(10*self._epsilon)

        self._H_bar = 0

        # NUTS run diagnostics
        # number of tree nodes created each NUTS iteration
        self._num_tree_node = 0

        # Create lists to store NUTS run diagnostics
        self._create_run_diagnostic_attributes()

    #=========================================================================
    #============================== Properties ===============================
    #=========================================================================
    @property
    def max_depth(self):
        return self._max_depth

    @max_depth.setter
    def max_depth(self, value):
        if value is None:
            value = 15 # default value
        if not isinstance(value, int):
            raise TypeError('max_depth must be an integer.')
        if value < 0:
            raise ValueError('max_depth must be >= 0.')
        self._max_depth = value

    @property
    def step_size(self):
        return self._step_size
    
    @step_size.setter
    def step_size(self, value):
        if value is None:
            pass # NUTS will adapt the step size

        # step_size must be a positive float, raise error otherwise
        elif isinstance(value, bool)\
            or not isinstance(value, Number)\
            or value <= 0:
            raise TypeError('step_size must be a positive float or None.')
        self._step_size = value

    @property
    def opt_acc_rate(self):
        return self._opt_acc_rate
    
    @opt_acc_rate.setter
    def opt_acc_rate(self, value):
        if not isinstance(value, Number) or value <= 0 or value >= 1:
            raise ValueError('opt_acc_rate must be a float in (0, 1).')
        self._opt_acc_rate = value

    #=========================================================================
    #================== Implement methods required by Sampler =============
    #=========================================================================
    def validate_target(self):
        # Check if the target has logd and gradient methods
        try:
            current_target_logd, current_target_grad =\
            self._nuts_target(np.ones(self.dim))
        except:
            raise ValueError('Target must have logd and gradient methods.')

    def reinitialize(self):
        # Call the parent reset method
        super().reinitialize()
        # Reset NUTS run diagnostic attributes
        self._reset_run_diagnostic_attributes()

    def step(self):
        if isinstance(self._epsilon_bar, str) and self._epsilon_bar == "unset":
            self._epsilon_bar = self._epsilon

        # Convert current_point, logd, and grad to numpy arrays
        # if they are CUQIarray objects
        if isinstance(self.current_point, CUQIarray):
            self.current_point = self.current_point.to_numpy() 
        if isinstance(self.current_target_logd, CUQIarray):
            self.current_target_logd = self.current_target_logd.to_numpy()
        if isinstance(self.current_target_grad, CUQIarray):
            self.current_target_grad = self.current_target_grad.to_numpy()

        # reset number of tree nodes for each iteration
        self._num_tree_node = 0

        # copy current point, logd, and grad in local variables
        point_k = self.current_point # initial position (parameters)
        logd_k = self.current_target_logd
        grad_k = self.current_target_grad # initial gradient
        
        # compute r_k and Hamiltonian
        r_k = self._Kfun(1, 'sample') # resample momentum vector
        Ham = logd_k - self._Kfun(r_k, 'eval') # Hamiltonian

        # slice variable
        log_u = Ham - np.random.exponential(1, size=1)

        # initialization
        j, s, n = 0, 1, 1
        point_minus, point_plus = point_k.copy(), point_k.copy()
        grad_minus, grad_plus = grad_k.copy(), grad_k.copy()
        r_minus, r_plus = r_k.copy(), r_k.copy()

        # run NUTS
        acc = 0
        while (s == 1) and (j <= self.max_depth):
            # sample a direction
            v = int(2*(np.random.rand() < 0.5)-1)

            # build tree: doubling procedure
            if (v == -1):
                point_minus, r_minus, grad_minus, _, _, _, \
                    point_prime, logd_prime, grad_prime,\
                        n_prime, s_prime, alpha, n_alpha = \
                            self._BuildTree(point_minus, r_minus, grad_minus,
                                            Ham, log_u, v, j, self._epsilon)
            else:
                _, _, _, point_plus, r_plus, grad_plus, \
                    point_prime, logd_prime, grad_prime,\
                        n_prime, s_prime, alpha, n_alpha = \
                            self._BuildTree(point_plus, r_plus, grad_plus,
                                            Ham, log_u, v, j, self._epsilon)

            # Metropolis step
            alpha2 = min(1, (n_prime/n)) #min(0, np.log(n_p) - np.log(n))
            if (s_prime == 1) and \
                (np.random.rand() <= alpha2) and \
                (not np.isnan(logd_prime)) and \
                (not np.isinf(logd_prime)):
                self.current_point = point_prime.copy()
                # copy if array, else assign if scalar
                self.current_target_logd = (
                        logd_prime.copy()
                        if isinstance(logd_prime, np.ndarray)
                        else logd_prime
                    )
                self.current_target_grad = grad_prime.copy()
                acc = 1


            # update number of particles, tree level, and stopping criterion
            n += n_prime
            dpoints = point_plus - point_minus
            s = s_prime *\
                int((dpoints @ r_minus.T) >= 0) * int((dpoints @ r_plus.T) >= 0)
            j += 1
            self._current_alpha_ratio = alpha/n_alpha

        # update run diagnostic attributes
        self._update_run_diagnostic_attributes(
            self._num_tree_node, self._epsilon, self._epsilon_bar)
        
        self._epsilon = self._epsilon_bar 
        if np.isnan(self.current_target_logd):
            raise NameError('NaN potential func')

        return acc

    def tune(self, skip_len, update_count):
        """ adapt epsilon during burn-in using dual averaging"""
        if isinstance(self._epsilon_bar, str) and self._epsilon_bar == "unset":
            self._epsilon_bar = 1

        k = update_count+1

        # Fixed parameters that do not change during the run
        gamma, t_0, kappa = 0.05, 10, 0.75 # kappa in (0.5, 1]

        eta1 = 1/(k + t_0)
        self._H_bar = (1-eta1)*self._H_bar +\
            eta1*(self.opt_acc_rate - (self._current_alpha_ratio))
        self._epsilon = np.exp(self._mu - (np.sqrt(k)/gamma)*self._H_bar)
        eta = k**(-kappa)
        self._epsilon_bar =\
            np.exp(eta*np.log(self._epsilon) +(1-eta)*np.log(self._epsilon_bar))

    #=========================================================================
    def _nuts_target(self, x): # returns logposterior tuple evaluation-gradient
        return self.target.logd(x), self.target.gradient(x)

    #=========================================================================
    # auxiliary standard Gaussian PDF: kinetic energy function
    # d_log_2pi = d*np.log(2*np.pi)
    def _Kfun(self, r, flag):
        if flag == 'eval': # evaluate
            return 0.5*(r.T @ r) #+ d_log_2pi 
        if flag == 'sample': # sample
            return np.random.standard_normal(size=self.dim)

    #=========================================================================
    def _FindGoodEpsilon(self, epsilon=1):
        point_k = self.current_point
        self.current_target_logd, self.current_target_grad = self._nuts_target(
            point_k)
        logd = self.current_target_logd
        grad = self.current_target_grad

        r = self._Kfun(1, 'sample')    # resample a momentum
        Ham = logd - self._Kfun(r, 'eval')     # initial Hamiltonian
        _, r_prime, logd_prime, grad_prime = self._Leapfrog(
            point_k, r, grad, epsilon)

        # trick to make sure the step is not huge, leading to infinite values of
        # the likelihood
        k = 1
        while np.isinf(logd_prime) or np.isinf(grad_prime).any():
            k *= 0.5
            _, r_prime, logd_prime, grad_prime = self._Leapfrog(
                point_k, r, grad, epsilon*k)
        epsilon = 0.5*k*epsilon

        # doubles/halves the value of epsilon until the accprob of the Langevin
        # proposal crosses 0.5
        Ham_prime = logd_prime - self._Kfun(r_prime, 'eval')
        log_ratio = Ham_prime - Ham
        a = 1 if log_ratio > np.log(0.5) else -1
        while (a*log_ratio > -a*np.log(2)):
            epsilon = (2**a)*epsilon
            _, r_prime, logd_prime, _ = self._Leapfrog(
                point_k, r, grad, epsilon)
            Ham_prime = logd_prime - self._Kfun(r_prime, 'eval')
            log_ratio = Ham_prime - Ham
        return epsilon

    #=========================================================================
    def _Leapfrog(self, point_old, r_old, grad_old, epsilon):
        # symplectic integrator: trajectories preserve phase space volumen
        r_new = r_old + 0.5*epsilon*grad_old     # half-step
        point_new = point_old + epsilon*r_new     # full-step
        logd_new, grad_new = self._nuts_target(point_new)     # new gradient
        r_new += 0.5*epsilon*grad_new     # half-step
        return point_new, r_new, logd_new, grad_new

    #=========================================================================
    def _BuildTree(
            self, point_k, r, grad, Ham, log_u, v, j, epsilon, Delta_max=1000):
        # Increment the number of tree nodes counter
        self._num_tree_node += 1

        if (j == 0):     # base case
            # single leapfrog step in the direction v
            point_prime, r_prime, logd_prime, grad_prime = self._Leapfrog(
                point_k, r, grad, v*epsilon)
            Ham_prime = logd_prime - self._Kfun(r_prime, 'eval') # Hamiltonian
                                                                 # eval
            n_prime = int(log_u <= Ham_prime)     # if particle is in the slice
            s_prime = int(log_u < Delta_max + Ham_prime)     # check U-turn
            #
            diff_Ham = Ham_prime - Ham

            # Compute the acceptance probability
            # alpha_prime = min(1, np.exp(diff_Ham))
            # written in a stable way to avoid overflow when computing
            # exp(diff_Ham) for large values of diff_Ham
            alpha_prime = 1 if diff_Ham > 0 else np.exp(diff_Ham)
            n_alpha_prime = 1
            #
            point_minus, point_plus = point_prime, point_prime
            r_minus, r_plus = r_prime, r_prime
            grad_minus, grad_plus = grad_prime, grad_prime
        else: 
            # recursion: build the left/right subtrees
            point_minus, r_minus, grad_minus, point_plus, r_plus, grad_plus, \
                point_prime, logd_prime, grad_prime,\
                    n_prime, s_prime, alpha_prime, n_alpha_prime = \
                        self._BuildTree(point_k, r, grad,
                                        Ham, log_u, v, j-1, epsilon)
            if (s_prime == 1): # do only if the stopping criteria does not 
                               # verify at the first subtree
                if (v == -1):
                    point_minus, r_minus, grad_minus, _, _, _, \
                        point_2prime, logd_2prime, grad_2prime,\
                            n_2prime, s_2prime, alpha_2prime, n_alpha_2prime = \
                                self._BuildTree(point_minus, r_minus, grad_minus,
                                                Ham, log_u, v, j-1, epsilon)
                else:
                    _, _, _, point_plus, r_plus, grad_plus, \
                        point_2prime, logd_2prime, grad_2prime,\
                            n_2prime, s_2prime, alpha_2prime, n_alpha_2prime = \
                                self._BuildTree(point_plus, r_plus, grad_plus,
                                                Ham, log_u, v, j-1, epsilon)

                # Metropolis step
                alpha2 = n_2prime / max(1, (n_prime + n_2prime))
                if (np.random.rand() <= alpha2):
                    point_prime = point_2prime.copy()
                    # copy if array, else assign if scalar
                    logd_prime = (
                        logd_2prime.copy()
                        if isinstance(logd_2prime, np.ndarray)
                        else logd_2prime
                    )
                    grad_prime = grad_2prime.copy()

                # update number of particles and stopping criterion
                alpha_prime += alpha_2prime
                n_alpha_prime += n_alpha_2prime
                dpoints = point_plus - point_minus
                s_prime = s_2prime *\
                    int((dpoints@r_minus.T)>=0) * int((dpoints@r_plus.T)>=0)
                n_prime += n_2prime

        return point_minus, r_minus, grad_minus, point_plus, r_plus, grad_plus,\
            point_prime, logd_prime, grad_prime,\
                n_prime, s_prime, alpha_prime, n_alpha_prime

    #=========================================================================
    #======================== Diagnostic methods =============================
    #=========================================================================

    def _create_run_diagnostic_attributes(self):
        """A method to create attributes to store NUTS run diagnostic."""
        self._reset_run_diagnostic_attributes()

    def _reset_run_diagnostic_attributes(self):
        """A method to reset attributes to store NUTS run diagnostic."""
        # List to store number of tree nodes created each NUTS iteration
        self.num_tree_node_list = []
        # List of step size used in each NUTS iteration 
        self.epsilon_list = []
        # List of burn-in step size suggestion during adaptation 
        # only used when adaptation is done
        # remains fixed after adaptation (after burn-in)
        self.epsilon_bar_list = []

    def _update_run_diagnostic_attributes(self, n_tree, eps, eps_bar):
        """A method to update attributes to store NUTS run diagnostic."""
        # Store the number of tree nodes created in iteration k
        self.num_tree_node_list.append(n_tree)
        # Store the step size used in iteration k
        self.epsilon_list.append(eps)
        # Store the step size suggestion during adaptation in iteration k
        self.epsilon_bar_list.append(eps_bar)


from cuqi.distribution import JointDistribution, Posterior
from cuqi.experimental.mcmc import Sampler
from cuqi.samples import Samples, JointSamples
from typing import Dict
import numpy as np
import warnings

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterable, **kwargs):
        warnings.warn("Module mcmc: tqdm not found. Install tqdm to get sampling progress.")
        return iterable

# Not subclassed from Sampler as Gibbs handles multiple samplers and samples multiple parameters
# Similar approach as for JointDistribution
class HybridGibbs: 
    """
    Hybrid Gibbs sampler for sampling a joint distribution.

    Gibbs sampling samples the variables of the distribution sequentially,
    one variable at a time. When a variable represents a random vector, the
    whole vector is sampled simultaneously.
    
    The sampling of each variable is done by sampling from the conditional
    distribution of that variable given the values of the other variables.
    This is often a very efficient way of sampling from a joint distribution
    if the conditional distributions are easy to sample from. 

    Hybrid Gibbs sampler is a generalization of the Gibbs sampler where the
    conditional distributions are sampled using different MCMC samplers.
    
    When the conditionals are sampled exactly, the samples from the Gibbs 
    sampler converge to the joint distribution. See e.g.
    Gelman et al. "Bayesian Data Analysis" (2014), Third Edition
    for more details.

    In each Gibbs step, the corresponding sampler state and history are stored, 
    then the sampler is reinitialized. After reinitialization, the sampler state
    and history are set back to the stored values. This ensures preserving the 
    statefulness of the samplers.

    The order in which the conditionals are sampled is the order of the
    variables in the sampling strategy, unless a different sampling order
    is specified by the parameter `scan_order`

    Parameters
    ----------
    target : cuqi.distribution.JointDistribution
        Target distribution to sample from.
    
    sampling_strategy : dict
        Dictionary of sampling strategies for each variable.
        Keys are variable names.
        Values are sampler objects.

    num_sampling_steps : dict, *optional*
        Dictionary of number of sampling steps for each variable.
        The sampling steps are defined as the number of times the sampler
        will call its step method in each Gibbs step.
        Default is 1 for all variables.

    scan_order : list or str, *optional*
        Order in which the conditional distributions are sampled.
        If set to "random", use a random ordering at each step.
        If not specified, it will be the order in the sampling_strategy.

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.

    Example
    -------
    .. code-block:: python

        import cuqi
        import numpy as np

        # Model and data
        A, y_obs, probinfo = cuqi.testproblem.Deconvolution1D(phantom='sinc').get_components()
        n = A.domain_dim

        # Define distributions
        d = cuqi.distribution.Gamma(1, 1e-4)
        l = cuqi.distribution.Gamma(1, 1e-4)
        x = cuqi.distribution.GMRF(np.zeros(n), lambda d: d)
        y = cuqi.distribution.Gaussian(A, lambda l: 1/l)

        # Combine into a joint distribution and create posterior
        joint = cuqi.distribution.JointDistribution(d, l, x, y)
        posterior = joint(y=y_obs)

        # Define sampling strategy
        sampling_strategy = {
            'x': cuqi.experimental.mcmc.LinearRTO(maxit=15),
            'd': cuqi.experimental.mcmc.Conjugate(),
            'l': cuqi.experimental.mcmc.Conjugate(),
        }

        # Define Gibbs sampler
        sampler = cuqi.experimental.mcmc.HybridGibbs(posterior, sampling_strategy)

        # Run sampler
        sampler.warmup(200)
        sampler.sample(1000)

        # Get samples removing burn-in
        samples = sampler.get_samples().burnthin(200)

        # Plot results
        samples['x'].plot_ci(exact=probinfo.exactSolution)
        samples['d'].plot_trace(figsize=(8,2))
        samples['l'].plot_trace(figsize=(8,2))
            
    """

    def __init__(self, target: JointDistribution, sampling_strategy: Dict[str, Sampler], num_sampling_steps: Dict[str, int] = None, scan_order = None, callback=None):

        # Store target and allow conditioning to reduce to a single density
        self.target = target() # Create a copy of target distribution (to avoid modifying the original)

        # Store sampler instances (again as a copy to avoid modifying the original)
        self.samplers = sampling_strategy.copy()

        # Store number of sampling steps for each parameter
        self.num_sampling_steps = num_sampling_steps

        # Store parameter names
        self.par_names = self.target.get_parameter_names()

        # Store the scan order
        self._scan_order = scan_order

        # Check that the parameters of the target align with the sampling_strategy and scan_order
        if set(self.par_names) != set(self.scan_order):
            raise ValueError("Parameter names in JointDistribution do not equal the names in the scan order.")

        # Initialize sampler (after target is set)
        self._initialize()

        # Set the callback function
        self.callback = callback

    def _initialize(self):
        """ Initialize sampler """

        # Initial points
        self.current_samples = self._get_initial_points()

        # Initialize sampling steps
        self._initialize_num_sampling_steps()

        # Allocate samples
        self._allocate_samples()

        # Set targets
        self._set_targets()

        # Initialize the samplers
        self._initialize_samplers()

        # Validate all targets for samplers.
        self.validate_targets()

    @property
    def scan_order(self):
        if self._scan_order is None:
            return list(self.samplers.keys())
        if self._scan_order == "random":
            arr = list(self.samplers.keys())
            np.random.shuffle(arr) # Shuffle works in-place
            return arr
        return self._scan_order

    # ------------ Public methods ------------
    def validate_targets(self):
        """ Validate each of the conditional targets used in the Gibbs steps """
        if not isinstance(self.target, (JointDistribution, Posterior)):
            raise ValueError('Target distribution must be a JointDistribution or Posterior.')
        for sampler in self.samplers.values():
            sampler.validate_target()

    def sample(self, Ns) -> 'HybridGibbs':
        """ Sample from the joint distribution using Gibbs sampling

        Parameters
        ----------
        Ns : int
            The number of samples to draw.

        """
        for idx in tqdm(range(Ns), "Sample: "):

            self.step()

            self._store_samples()

            # Call callback function if specified
            self._call_callback(idx, Ns)

        return self

    def warmup(self, Nb, tune_freq=0.1) -> 'HybridGibbs':
        """ Warmup (tune) the samplers in the Gibbs sampling scheme

        Parameters
        ----------
        Nb : int
            The number of samples to draw during warmup.

        tune_freq : float, optional
            Frequency of tuning the samplers. Tuning is performed every tune_freq*Nb steps.

        """

        tune_interval = max(int(tune_freq * Nb), 1)

        for idx in tqdm(range(Nb), "Warmup: "):

            self.step()

            # Tune the sampler at tuning intervals (matching behavior of Sampler class)
            if (idx + 1) % tune_interval == 0:
                self.tune(tune_interval, idx // tune_interval) 
                
            self._store_samples()

            # Call callback function if specified
            self._call_callback(idx, Nb)

        return self

    def get_samples(self) -> Dict[str, Samples]:
        samples_object = JointSamples()
        for par_name in self.par_names:
            samples_array = np.array(self.samples[par_name]).T
            samples_object[par_name] = Samples(samples_array, self.target.get_density(par_name).geometry)
        return samples_object
    
    def step(self):
        """ Sequentially go through all parameters and sample them conditionally on each other """

        # Sample from each conditional distribution
        for par_name in self.scan_order:

            # Set target for current parameter
            self._set_target(par_name)

            # Get sampler
            sampler = self.samplers[par_name]

            # Instead of simply changing the target of the sampler, we reinitialize it.
            # This is to ensure that all internal variables are set to match the new target.
            # To return the sampler to the old state and history, we first extract the state and history
            # before reinitializing the sampler and then set the state and history back to the sampler

            # Extract state and history from sampler
            sampler_state = sampler.get_state()
            sampler_history = sampler.get_history()

            # Reinitialize sampler
            sampler.reinitialize()

            # Set state and history back to sampler
            sampler.set_state(sampler_state)
            sampler.set_history(sampler_history)

            # Allow for multiple sampling steps in each Gibbs step
            for _ in range(self.num_sampling_steps[par_name]):
                # Sampling step
                acc = sampler.step()

                # Store acceptance rate in sampler (matching behavior of Sampler class Sample method)
                sampler._acc.append(acc)

            # Extract samples (Ensure even 1-dimensional samples are 1D arrays)
            if isinstance(sampler.current_point, np.ndarray):
                self.current_samples[par_name] = sampler.current_point.reshape(-1)
            else:
                self.current_samples[par_name] = sampler.current_point

    def tune(self, skip_len, update_count):
        """ Run a single tuning step on each of the samplers in the Gibbs sampling scheme

        Parameters
        ----------
        skip_len : int
            Defines the number of steps in between tuning (i.e. the tuning interval).

        update_count : int
            The number of times tuning has been performed. Can be used for internal bookkeeping.

        """
        for par_name in self.par_names:
            self.samplers[par_name].tune(skip_len=skip_len, update_count=update_count)

    # ------------ Private methods ------------
    def _call_callback(self, sample_index, num_of_samples):
        """ Calls the callback function. Assumes input is sampler, sample index, and total number of samples """
        if self.callback is not None:
            self.callback(self, sample_index, num_of_samples)

    def _initialize_samplers(self):
        """ Initialize samplers """
        for sampler in self.samplers.values():
            sampler.initialize()

    def _initialize_num_sampling_steps(self):
        """ Initialize the number of sampling steps for each sampler. Defaults to 1 if not set by user """

        if self.num_sampling_steps is None:
            self.num_sampling_steps = {par_name: 1 for par_name in self.par_names}

        for par_name in self.par_names:
            if par_name not in self.num_sampling_steps:
                self.num_sampling_steps[par_name] = 1


    def _set_targets(self):
        """ Set targets for all samplers using the current samples """
        par_names = self.par_names
        for par_name in par_names:
            self._set_target(par_name)

    def _set_target(self, par_name):
        """ Set target conditional distribution for a single parameter using the current samples """
        # Get all other conditional parameters other than the current parameter and update the target
        # This defines - from a joint p(x,y,z) - the conditional distribution p(x|y,z) or p(y|x,z) or p(z|x,y)
        conditional_params = {par_name_: self.current_samples[par_name_] for par_name_ in self.par_names if par_name_ != par_name}
        self.samplers[par_name].target = self.target(**conditional_params)

    def _allocate_samples(self):
        """ Allocate memory for samples """
        samples = {}
        for par_name in self.par_names:
            samples[par_name] = []
        self.samples = samples

    def _get_initial_points(self):
        """ Get initial points for each parameter """
        initial_points = {}
        for par_name in self.par_names:
            sampler = self.samplers[par_name]
            if sampler.initial_point is None:
                sampler.initial_point = sampler._get_default_initial_point(self.target.get_density(par_name).dim)
            initial_points[par_name] = sampler.initial_point
            
        return initial_points

    def _store_samples(self):
        """ Store current samples at index i of samples dict """
        for par_name in self.par_names:
            self.samples[par_name].append(self.current_samples[par_name])

    def __repr__(self):
        """ Return a string representation of the sampler. """
        msg = f"Sampler: {self.__class__.__name__} \n"
        if self.target is None:
            msg += f" Target: None \n"
        else:
            msg += f" Target: \n \t {self.target} \n\n"
            
        for key, value in zip(self.samplers.keys(), self.samplers.values()):
            msg += f" Variable '{key}' with {value} \n"

        return msg



import numpy as np
import scipy as sp

from cuqi.solver import ScipyLBFGSB, ScipyMinimizer, ScipyLSQ, ScipyLinearLSQ, CGLS, LM, FISTA, ADMM, ProximalL1, ProjectNonnegative, ProjectBox
from scipy.optimize import lsq_linear


def test_CGLS():
    # Parameters
    rng = np.random.default_rng()
    m, n = 1000, 500
    A = sp.sparse.rand(m, n, density=1e-4, random_state=rng)
    b = rng.standard_normal(m)
    res = lsq_linear(A, b, tol=1e-8)
    ref_sol = res.x
    #
    x0 = np.zeros(n)
    sol, _ = CGLS(A, b, x0, int(1e4), tol=1e-8).solve()

    # Compare
    assert np.allclose(sol, ref_sol, rtol=1e-3)

def test_ScipyMinimizer_without_gradient():
    solver = ScipyMinimizer(sp.optimize.rosen,
                            np.array([1.3, 0.7, 0.8, 1.9, 1.2]),
                            method='Nelder-Mead',
                            tol=1e-6)
    sol, _ = solver.solve()
    sol_ref = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    assert np.allclose(sol, sol_ref)

def test_ScipyMinimizer_with_gradient():
    solver = ScipyMinimizer(sp.optimize.rosen,
                            np.array([1.3, 0.7, 0.8, 1.9, 1.2]),
                            gradfunc=sp.optimize.rosen_der,
                            method='BFGS',
                            tol=1e-6)
    sol, _ = solver.solve()
    sol_ref = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    assert np.allclose(sol, sol_ref)

def test_ScipyLBFGSB_without_gradient():
    solver = ScipyLBFGSB(sp.optimize.rosen,
                         np.array([1.3, 0.7, 0.8, 1.9, 1.2]))
    sol, _ = solver.solve()
    sol_ref = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    assert np.allclose(sol, sol_ref)

def test_ScipyLBFGSB_with_gradient():
    solver = ScipyLBFGSB(sp.optimize.rosen,
                         np.array([1.3, 0.7, 0.8, 1.9, 1.2]),
                         gradfunc=sp.optimize.rosen_der)
    sol, _ = solver.solve()
    sol_ref = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    assert np.allclose(sol, sol_ref)

def test_ScipyLSQ_without_Jac():
    def fun_rosenbrock(x):
        return np.array([10 * (x[1] - x[0]**2), (1 - x[0])])
    x0 = np.array([2, 2])
    solver = ScipyLSQ(fun_rosenbrock, x0)
    sol, _ = solver.solve()
    sol_ref = np.array([1, 1])
    assert np.allclose(sol, sol_ref)

def test_ScipyLSQ_with_Jac():
    def fun_rosenbrock(x):
        return np.array([10 * (x[1] - x[0]**2), (1 - x[0])])
    def jac_rosenbrock(x):
        return np.array([
            [-20 * x[0], 10],
            [-1, 0]])
    x0 = np.array([2, 2])
    solver = ScipyLSQ(fun_rosenbrock, x0, jacfun=jac_rosenbrock)
    sol, _ = solver.solve()
    sol_ref = np.array([1, 1])
    assert np.allclose(sol, sol_ref)

def test_ScipyLinearLSQ_with_matrix():
    rng = np.random.default_rng(seed = 1219)
    m, n = 10, 5
    A = rng.standard_normal((m, n))
    b = rng.standard_normal(m)
    res = lsq_linear(A, b, tol=1e-8)
    ref_sol = res.x
    sol, _ = ScipyLinearLSQ(A, b).solve()
    assert np.allclose(sol, ref_sol, rtol=1e-10)

def test_ScipyLinearLSQ_with_LinearOperator():
    rng = np.random.default_rng(seed = 1219)
    m, n = 10, 5
    A = rng.standard_normal((m, n))
    b = rng.standard_normal(m)
    A_op = sp.sparse.linalg.LinearOperator((m, n),
                                           matvec=lambda x: A @ x,
                                           rmatvec=lambda x: A.T @ x
                                           )
    res = lsq_linear(A, b, tol=1e-8)
    ref_sol = res.x
    sol, _ = ScipyLinearLSQ(A_op, b).solve()
    assert np.allclose(sol, ref_sol, rtol=1e-10)

def test_ScipyLinearLSQ_against_ScipyMinimizer_and_against_FISTA():
    A = np.array([[73,71,52],[87,74,46],[72,2,7],[80,89,71]])
    b = np.array([49,67,68,20])

    # solve with ScipyMinimizer
    def fun(x):
        return 0.5*np.linalg.norm(A@x-b)**2
    def jac(x):
        return A.T@(A@x-b)
    sol_min, _ = ScipyMinimizer(fun, np.zeros(3), gradfunc=jac, tol=1e-10, bounds=[(0,np.inf),(0,np.inf),(0,np.inf)]).solve()

    # solve with ScipyLinearLSQ
    sol_lsq, _ = ScipyLinearLSQ(A, b, ([0,0,0],[np.inf,np.inf,np.inf]), tol=1e-10).solve()

    # solve with FISTA
    rng = np.random.default_rng(seed = 1219)
    x0 = rng.standard_normal(3)
    sol_fista, _ = FISTA(A, b, lambda x, _: ProjectNonnegative(x), x0, stepsize=1e-7, maxit=100000, abstol=1e-16, adaptive=True).solve()

    assert np.allclose(sol_lsq, sol_fista, rtol=1e-8)
    assert np.allclose(sol_min, sol_lsq, rtol=1e-8)

def test_LM():
    # compare to MATLAB's original code solution
    t = np.arange(1, 10, 2)
    A = lambda x: x[0]*(1-np.exp(-x[1]*t))
    x_true = np.array([1, 0.1]) #  true parameter values
    N = len(x_true)
    #
    b = np.array([0.0886724443121281, 0.270993439737938, 0.38588480731453, 0.492318565823575, 0.584974827859323])
    M = len(b)
    #    
    lam = 1
    x0 = np.ones(N)
    Q = np.eye(M)
    residual = lambda x: Q.T @ (np.sqrt(lam)*(A(x)-b))
    jacobian = lambda x: Q.T @ (np.sqrt(lam)*np.array([1-np.exp(-x[1]*t), x[0]*t*np.exp(-x[1]*t)]).T)
    opt, info = LM(residual, x0, jacobian, maxit=int(1e4), tol=1e-3, gradtol=1e-8, sparse=False).solve()
    MSE = np.linalg.norm(info['func'])**2/(M-N)

    # solution TwoParameters.m from John's book
    ref_MSE = 8.06826104975311e-05
    ref_opt = np.array([0.92953293909007, 0.109186388510998])

    assert np.allclose(MSE, ref_MSE) and np.allclose(opt, ref_opt)
    
def test_FISTA():
    # Parameters
    rng = np.random.default_rng(seed = 42)
    m, n = 10, 5
    A = rng.standard_normal((m, n))
    b = rng.standard_normal(m)
    
    stepsize = 0.99/(sp.linalg.interpolative.estimate_spectral_norm(A)**2)
    x0 = np.zeros(n)
    sol, _ = FISTA(A, b, ProximalL1, x0, stepsize = stepsize, maxit = 100, abstol=1e-12, adaptive = True).solve()

    ref_sol = np.array([-1.83273787e-03, -1.72094582e-13,  0.0, -3.35835639e-01, -1.27795593e-01])
    # Compare
    assert np.allclose(sol, ref_sol, atol=1e-4)

def test_ADMM_matrix_form():
    # Parameters
    rng = np.random.default_rng(seed = 42)
    m, n = 10, 5
    A = rng.standard_normal((m, n))
    b = rng.standard_normal(m)
    
    k = 4
    L = rng.standard_normal((k, n))

    x0 = np.zeros(n)
    sol, _ = ADMM(A, b, [(ProximalL1, np.eye(n)), (lambda z, _ : ProjectNonnegative(z), L)],
                   x0, 10, maxit = 100, adaptive = True).solve()

    ref_sol = np.array([-3.99513417e-03, -1.32339656e-01, -4.52822633e-02, -7.44973888e-02, -3.35005208e-11])
    # Compare
    assert np.allclose(sol, ref_sol, atol=1e-4)


def test_ADMM_function_form():
    # Parameters
    rng = np.random.default_rng(seed = 42)
    m, n = 10, 5
    A = rng.standard_normal((m, n))
    def A_fun(x, flag):
        if flag == 1:
            return A@x
        if flag == 2:
            return A.T@x
        
    b = rng.standard_normal(m)
    
    k = 4
    L = rng.standard_normal((k, n))

    x0 = np.zeros(n)
    sol, _ = ADMM(A_fun, b, [(ProximalL1, np.eye(n)), (lambda z, _ : ProjectNonnegative(z), L)],
                   x0, 10, maxit = 100, adaptive = True).solve()

    print(sol)
    ref_sol = np.array([-3.99513417e-03, -1.32339656e-01, -4.52822633e-02, -7.44973888e-02, -3.35005208e-11])
    # Compare
    assert np.allclose(sol, ref_sol, atol=1e-4)



from ._solver import (
    ScipyLBFGSB,
    ScipyMinimizer,
    ScipyMaximizer,
    ScipyLSQ,
    ScipyLinearLSQ,
    CGLS,
    LM,
    PDHG,
    FISTA,
    ADMM,
    ProjectNonnegative,
    ProjectBox,
    ProximalL1
)



import numpy as np
import numbers
from scipy.special import erf
from cuqi.geometry import _get_identity_geometries
from cuqi.utilities import force_ndarray
from cuqi.distribution import Distribution

class Normal(Distribution):
    """
    Normal probability distribution. Generates instance of cuqi.distribution.Normal. The variables of this distribution are iid.

    
    Parameters
    ------------
    mean: mean of distribution
    std: standard deviation
    
    Example
    -----------
    .. code-block:: python

        #Generate Normal with mean 2 and standard deviation 1
        p = cuqi.distribution.Normal(mean=2, std=1)
    """
    def __init__(self, mean=None, std=None, is_symmetric=True, **kwargs):
        # Init from abstract distribution class
        super().__init__(is_symmetric=is_symmetric, **kwargs)  

        # Init specific to this distribution
        self.mean = mean
        self.std = std

    @property
    def mean(self):
        """ Mean of the distribution """
        return self._mean

    @mean.setter
    def mean(self, value):
        self._mean = force_ndarray(value, flatten=True)

    @property
    def std(self):
        """ Std of the distribution """
        return self._std

    @std.setter
    def std(self, value):
        self._std = force_ndarray(value, flatten=True)

    def pdf(self, x):
        return np.prod(1/(self.std*np.sqrt(2*np.pi))*np.exp(-0.5*((x-self.mean)/self.std)**2))

    def logpdf(self, x):
        return np.sum(-np.log(self.std*np.sqrt(2*np.pi))-0.5*((x-self.mean)/self.std)**2)

    def cdf(self, x):
        return np.prod(0.5*(1 + erf((x-self.mean)/(self.std*np.sqrt(2)))))

    def _gradient(self, val, *args, **kwargs):
        if not type(self.geometry) in _get_identity_geometries():
            raise NotImplementedError("Gradient not implemented for distribution {} with geometry {}".format(self,self.geometry))
        if not callable(self.mean):
            return -(val-self.mean)/(self.std**2)
        elif hasattr(self.mean, "gradient"): # for likelihood
            model = self.mean
            dev = val - model.forward(*args, **kwargs)
            print(dev)
            return model.gradient(1.0/(np.array(self.std)) @ dev, *args, **kwargs)
        else:
            raise NotImplementedError("Gradient not implemented for distribution {} with location {}".format(self,self.mean))

    def _sample(self,N=1, rng=None):

        """
        Draw sample(s) from distribution
        
        Example
        -------
        p = cuqi.distribution.Normal(mean=2, std=1) #Define distribution
        s = p.sample() #Sample from distribution
        

        Returns
        -------
        Generated sample(s)

        """

        if rng is not None:
            s =  rng.normal(self.mean, self.std, (N,self.dim)).T
        else:
            s = np.random.normal(self.mean, self.std, (N,self.dim)).T
        return s


from cuqi.implicitprior import RegularizedGaussian
from cuqi.distribution import Distribution, Gaussian

import numpy as np


class RegularizedUnboundedUniform(RegularizedGaussian):
        """ Implicit Regularized Unbounded Uniform. 

        Defines a so-called implicit prior with implicit regularization on a Gaussian distribution with zero precision.
        The regularization can be defined in the form of a proximal operator or a projector.
        Alternatively, preset constraints and regularization can be used.

        For regularization of the form f(x), provide a single proximal operator.

        Can be used as a prior in a posterior which can be sampled with the RegularizedLinearRTO sampler.

        Alias for :class:`~cuqi.implicitprior.RegularizedGaussian` with zero mean and zero sqrtprec.

        For more details on implicit regularized Gaussian see the following paper:

        [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized
        Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

        Parameters
        ----------
        geometry : :class:`~cuqi.geometry.Geometry` or integer
                The geometry of the underlying variable. Defines the dimension of the distribution. 

        proximal : callable f(x, scale) or None
                Euclidean proximal operator f of the regularization function g, that is, a solver for the optimization problem
                min_z 0.5||x-z||_2^2+scale*g(x).

        projector : callable f(x) or None
                Euclidean projection onto the constraint C, that is, a solver for the optimization problem
                min_(z in C) 0.5||x-z||_2^2.

        constraint : string or None
                Preset constraints. Can be set to "nonnegativity" and "box". Required for use in Gibbs.
                For "box", the following additional parameters can be passed:
                lower_bound : array_like or None
                        Lower bound of box, defaults to zero
                upper_bound : array_like
                        Upper bound of box, defaults to one

        regularization : string or None
                Preset regularization. Can be set to "l1". Required for use in Gibbs in future update.
                For "l1" or "tv", the following additional parameters can be passed:
                strength : scalar
                        Regularization parameter, i.e., strength*||Lx||_1, defaults to one

        """
        def __init__(self, geometry, proximal = None, projector = None, constraint = None, regularization = None, **kwargs):
                
                args = {"lower_bound" : kwargs.pop("lower_bound", None),
                        "upper_bound" : kwargs.pop("upper_bound", None),
                        "strength" : kwargs.pop("strength", None)}

                # Underlying explicit Gaussian
                # This line throws a warning due trying to applying get_sqrtprec_from_sqrtprec to an all zero matrix  
                self._gaussian = Gaussian(mean = np.zeros(geometry.par_dim), sqrtprec = np.zeros((geometry.par_dim,geometry.par_dim)), geometry = geometry, **kwargs) 

                # Init from abstract distribution class
                super(Distribution, self).__init__(**kwargs)

                self._force_list = False
                self._parse_regularization_input_arguments(proximal, projector, constraint, regularization, args)



import numpy as np
import cuqi
from cuqi.experimental.mcmc import Sampler
from cuqi.array import CUQIarray

class PCN(Sampler):  # Refactor to Proposal-based sampler?

    _STATE_KEYS = Sampler._STATE_KEYS.union({'scale', 'current_likelihood_logd', 'lambd'})

    def __init__(self, target=None, scale=1.0, **kwargs):

        super().__init__(target, **kwargs)
        self.initial_scale = scale

    def _initialize(self):
        self.scale = self.initial_scale
        self.current_likelihood_logd = self._loglikelihood(self.current_point)

        # parameters used in the Robbins-Monro recursion for tuning the scale parameter
        # see details and reference in the tune method
        self.lambd = self.scale
        self.star_acc = 0.44 #TODO: 0.234 # target acceptance rate

    def validate_target(self):
        if not isinstance(self.target, cuqi.distribution.Posterior):
            raise ValueError(f"To initialize an object of type {self.__class__}, 'target' need to be of type 'cuqi.distribution.Posterior'.")
        if not isinstance(self.prior, (cuqi.distribution.Gaussian, cuqi.distribution.Normal)):
            raise ValueError("The prior distribution of the target need to be Gaussian")

    def step(self):
        # propose state
        xi = self.prior.sample(1).flatten()   # sample from the prior
        x_star = np.sqrt(1-self.scale**2)*self.current_point + self.scale*xi   # PCN proposal

        # evaluate target
        loglike_eval_star =  self._loglikelihood(x_star) 

        # ratio and acceptance probability
        ratio = loglike_eval_star - self.current_likelihood_logd  # proposal is symmetric
        alpha = min(0, ratio)

        # accept/reject
        acc = 0
        u_theta = np.log(np.random.rand())
        if (u_theta <= alpha):
            self.current_point = x_star
            self.current_likelihood_logd = loglike_eval_star
            acc = 1
        
        return acc

    @property
    def prior(self):
        return self.target.prior

    @property
    def likelihood(self):
        return self.target.likelihood
        
    def _loglikelihood(self, x):
        return self.likelihood.logd(x)

    @property
    def dim(self): # TODO. Check if we need this. Implemented in base class
        if hasattr(self,'target') and hasattr(self.target,'dim'):
            self._dim = self.target.dim
        elif hasattr(self,'target') and isinstance(self.target,tuple) and len(self.target)==2:
            self._dim = self.target[0].dim
        return self._dim

    def tune(self, skip_len, update_count):
        """
        Tune the scale parameter of the PCN sampler.
        The tuning is based on algorithm 4 in Andrieu, Christophe, and Johannes Thoms. 
        "A tutorial on adaptive MCMC." Statistics and computing 18 (2008): 343-373.
        Note: the tuning algorithm here is the same as the one used in MH sampler.
        """

        # average acceptance rate in the past skip_len iterations
        hat_acc = np.mean(self._acc[-skip_len:])

        # new scaling parameter zeta to be used in the Robbins-Monro recursion
        zeta = 1/np.sqrt(update_count+1)

        # Robbins-Monro recursion to ensure that the variation of lambd vanishes
        self.lambd = np.exp(np.log(self.lambd) + zeta*(hat_acc-self.star_acc))

        # update scale parameter
        self.scale = min(self.lambd, 1)



import numpy as np
import cuqi
from cuqi.experimental.mcmc import Sampler
from cuqi.implicitprior import RestorationPrior, MoreauYoshidaPrior
from cuqi.array import CUQIarray
from copy import copy

class ULA(Sampler): # Refactor to Proposal-based sampler?
    """Unadjusted Langevin algorithm (ULA) (Roberts and Tweedie, 1996)

    It approximately samples a distribution given its logpdf gradient based on
    the Langevin diffusion dL_t = dW_t + 1/2*Nabla target.logd(L_t)dt, where
    W_t is the `dim`-dimensional standard Brownian motion.
    ULA results from the Euler-Maruyama discretization of this Langevin stochastic
    differential equation (SDE). 

    For more details see: Roberts, G. O., & Tweedie, R. L. (1996). Exponential convergence
    of Langevin distributions and their discrete approximations. Bernoulli, 341-363.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution`
        The target distribution to sample. Must have logd and gradient method. Custom logpdfs 
        and gradients are supported by using a :class:`cuqi.distribution.UserDefinedDistribution`.
    
    initial_point : ndarray
        Initial parameters. *Optional*

    scale : float
        The Langevin diffusion discretization time step (In practice, scale must
        be smaller than 1/L, where L is the Lipschitz of the gradient of the log
        target density, logd).

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.


    Example
    -------
    .. code-block:: python

        # Parameters
        dim = 5 # Dimension of distribution
        mu = np.arange(dim) # Mean of Gaussian
        std = 1 # standard deviation of Gaussian

        # Logpdf function
        logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)
        gradient_func = lambda x: -2/(std**2)*(x - mu)

        # Define distribution from logpdf and gradient as UserDefinedDistribution
        target = cuqi.distribution.UserDefinedDistribution(dim=dim, logpdf_func=logpdf_func,
            gradient_func=gradient_func)

        # Set up sampler
        sampler = cuqi.experimental.mcmc.ULA(target, scale=1/dim**2)

        # Sample
        sampler.sample(2000)

    A Deblur example can be found in demos/demo27_ULA.py
    # TODO: update demo once sampler merged
    """

    _STATE_KEYS = Sampler._STATE_KEYS.union({'scale', 'current_target_grad'})

    def __init__(self, target=None, scale=1.0, **kwargs):

        super().__init__(target, **kwargs)
        self.initial_scale = scale

    def _initialize(self):
        self.scale = self.initial_scale
        self.current_target_grad = self._eval_target_grad(self.current_point)

    def validate_target(self):
        try:
            self._eval_target_grad(np.ones(self.dim))
            pass
        except (NotImplementedError, AttributeError):
            raise ValueError("The target needs to have a gradient method")

    def _eval_target_logd(self, x):
        return None

    def _eval_target_grad(self, x):
        return self.target.gradient(x)

    def _accept_or_reject(self, x_star, target_eval_star, target_grad_star):
        """
        Accepts the proposed state and updates the sampler's state accordingly, i.e.,
        current_point, current_target_eval, and current_target_grad_eval.

        Parameters
        ----------
        x_star : 
            The proposed state

        target_eval_star: 
            The log likelihood evaluated at x_star

        target_grad_star: 
            The gradient of log likelihood evaluated at x_star

        Returns
        -------
        scalar
            1 (accepted)
        """

        self.current_point = x_star
        self.current_target_grad = target_grad_star
        acc = 1

        return acc

    def step(self):
        # propose state
        xi = cuqi.distribution.Normal(mean=np.zeros(self.dim), std=np.sqrt(self.scale)).sample()
        x_star = self.current_point + 0.5*self.scale*self.current_target_grad + xi

        # evaluate target
        target_eval_star = self._eval_target_logd(x_star)
        target_grad_star = self._eval_target_grad(x_star)

        # accept or reject proposal
        acc = self._accept_or_reject(x_star, target_eval_star, target_grad_star)

        return acc

    def tune(self, skip_len, update_count):
        pass


class MALA(ULA): # Refactor to Proposal-based sampler?
    """  Metropolis-adjusted Langevin algorithm (MALA) (Roberts and Tweedie, 1996)

    Samples a distribution given its logd and gradient (up to a constant) based on
    Langevin diffusion dL_t = dW_t + 1/2*Nabla target.logd(L_t)dt,
    W_t is the `dim`-dimensional standard Brownian motion.
    A sample is firstly proposed by ULA and is then accepted or rejected according
    to a Metropolis–Hastings step.
    This accept-reject step allows us to remove the asymptotic bias of ULA.

    For more details see: Roberts, G. O., & Tweedie, R. L. (1996). Exponential convergence
    of Langevin distributions and their discrete approximations. Bernoulli, 341-363.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution`
        The target distribution to sample. Must have logpdf and gradient method. Custom logpdfs 
        and gradients are supported by using a :class:`cuqi.distribution.UserDefinedDistribution`.
    
    initial_point : ndarray
        Initial parameters. *Optional*

    scale : float
        The Langevin diffusion discretization time step (In practice, scale must
        be smaller than 1/L, where L is the Lipschitz of the gradient of the log
        target density, logd).

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.


    Example
    -------
    .. code-block:: python

        # Parameters
        dim = 5 # Dimension of distribution
        mu = np.arange(dim) # Mean of Gaussian
        std = 1 # standard deviation of Gaussian

        # Logpdf function
        logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)
        gradient_func = lambda x: -2/(std**2)*(x-mu)

        # Define distribution from logpdf as UserDefinedDistribution (sample and gradients also supported)
        target = cuqi.distribution.UserDefinedDistribution(dim=dim, logpdf_func=logpdf_func,
            gradient_func=gradient_func)

        # Set up sampler
        sampler = cuqi.experimental.mcmc.MALA(target, scale=1/5**2)

        # Sample
        sampler.sample(2000)

    A Deblur example can be found in demos/demo28_MALA.py
    # TODO: update demo once sampler merged
    """

    _STATE_KEYS = ULA._STATE_KEYS.union({'current_target_logd'})

    def _initialize(self):
        super()._initialize()
        self.current_target_logd = self.target.logd(self.current_point)

    def _eval_target_logd(self, x):
        return self.target.logd(x)

    def _accept_or_reject(self, x_star, target_eval_star, target_grad_star):
        """
        Accepts the proposed state according to a Metropolis step and updates
        the sampler's state accordingly, i.e., current_point, current_target_eval,
        and current_target_grad_eval.

        Parameters
        ----------
        x_star : 
            The proposed state

        target_eval_star: 
            The log likelihood evaluated at x_star

        target_grad_star: 
            The gradient of log likelihood evaluated at x_star

        Returns
        -------
        scaler
            1 if accepted, 0 otherwise
        """
        log_target_ratio = target_eval_star - self.current_target_logd
        log_prop_ratio = self._log_proposal(self.current_point, x_star, target_grad_star) \
            - self._log_proposal(x_star, self.current_point,  self.current_target_grad)
        log_alpha = min(0, log_target_ratio + log_prop_ratio)

        # accept/reject with Metropolis
        acc = 0
        log_u = np.log(np.random.rand())
        if (log_u <= log_alpha) and \
           (not np.isnan(target_eval_star)) and \
           (not np.isinf(target_eval_star)):
            self.current_point = x_star
            self.current_target_logd = target_eval_star
            self.current_target_grad = target_grad_star
            acc = 1
        return acc

    def tune(self, skip_len, update_count):
        pass

    def _log_proposal(self, theta_star, theta_k, g_logpi_k):
        mu = theta_k + ((self.scale)/2)*g_logpi_k
        misfit = theta_star - mu
        return -0.5*((1/(self.scale))*(misfit.T @ misfit))


class MYULA(ULA):
    """Moreau-Yoshida Unadjusted Langevin algorithm (MYUULA) (Durmus et al., 2018)

    Samples a smoothed target distribution given its smoothed logpdf gradient.
    It is based on the Langevin diffusion dL_t = dW_t + 1/2*Nabla target.logd(L_t)dt, 
    where W_t is a `dim`-dimensional standard Brownian motion.
    It targets a differentiable density (partially) smoothed by the Moreau-Yoshida
    envelope. The smoothed target density can be made arbitrarily closed to the
    true unsmoothed target density.

    For more details see: Durmus, Alain, Eric Moulines, and Marcelo Pereyra.
    "Efficient Bayesian
    computation by proximal Markov chain Monte Carlo: when Langevin meets Moreau."
    SIAM Journal on Imaging Sciences 11.1 (2018): 473-506.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution`
        The target distribution to sample from. The target distribution results from
        a differentiable likelihood and prior of type RestorationPrior.
    
    initial_point : ndarray
        Initial parameters. *Optional*

    scale : float
        The Langevin diffusion discretization time step (In practice, scale must
        be smaller than 1/L, where L is the Lipschitz of the gradient of the log
        target density, logd).
        
    smoothing_strength : float
        This parameter controls the smoothing strength of MYULA.

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.

    A Deblur example can be found in demos/howtos/myula.py
    # TODO: update demo once sampler merged
    """
    def __init__(self, target=None, scale=1.0, smoothing_strength=0.1, **kwargs):
        self.smoothing_strength = smoothing_strength
        super().__init__(target=target, scale=scale, **kwargs)

    @Sampler.target.setter
    def target(self, value):
        """ Set the target density. Runs validation of the target. """
        self._target = value

        if self._target is not None:
            # Create a smoothed target
            self._smoothed_target = self._create_smoothed_target(value)

            # Validate the target
            self.validate_target()

    def _create_smoothed_target(self, value):
        """ Create a smoothed target using a Moreau-Yoshida envelope. """
        copied_value = copy(value)
        if isinstance(copied_value.prior, RestorationPrior):
            # Acceess the prior name
            name = value.prior.name
            copied_value.prior = MoreauYoshidaPrior(
                copied_value.prior,
                self.smoothing_strength,
                name=name)
        return copied_value

    def validate_target(self):
        # Call ULA target validation
        super().validate_target()

        # Additional validation for MYULA target
        if isinstance(self.target.prior, MoreauYoshidaPrior):
            raise ValueError(("The prior is already smoothed, apply"
                              " ULA when using a MoreauYoshidaPrior."))
        if not hasattr(self.target.prior, "restore"):
            raise NotImplementedError(
                ("Using MYULA with a prior that does not have a restore method"
                " is not supported.")
            )

    def _eval_target_grad(self, x):
        return self._smoothed_target.gradient(x)

class PnPULA(MYULA):
    """Plug-and-Play Unadjusted Langevin algorithm (PnP-ULA)
    (Laumont et al., 2022)

    Samples a smoothed target distribution given its smoothed logpdf gradient based on
    Langevin diffusion dL_t = dW_t + 1/2*Nabla target.logd(L_t)dt, where W_t is
    a `dim`-dimensional standard Brownian motion.
    It targets a differentiable density (partially) smoothed by a convolution
    with Gaussian kernel with zero mean and smoothing_strength variance. The
    smoothed target density can be made arbitrarily closed to the
    true unsmoothed target density. 

    For more details see: Laumont, R., Bortoli, V. D., Almansa, A., Delon, J.,
    Durmus, A., & Pereyra, M. (2022). Bayesian imaging using plug & play priors:
    when Langevin meets Tweedie. SIAM Journal on Imaging Sciences, 15(2), 701-737.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution`
        The target distribution to sample. The target distribution result from
        a differentiable likelihood and prior of type RestorationPrior.
    
    initial_point : ndarray
        Initial parameters. *Optional*

    scale : float
        The Langevin diffusion discretization time step (In practice, a scale of
        1/L, where L is the Lipschitz of the gradient of the log target density
        is recommended but not guaranteed to be the optimal choice).
        
    smoothing_strength : float
        This parameter controls the smoothing strength of PnP-ULA.


    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.

    # TODO: update demo once sampler merged
    """
    def __init__ (self, target=None, scale=1.0, smoothing_strength=0.1, **kwargs):
        super().__init__(target=target, scale=scale, 
                         smoothing_strength=smoothing_strength, **kwargs)



from __future__ import annotations
from typing import List, Any, Union
from ._ast import VariableNode, Node
from ._orderedset import _OrderedSet
import operator
import cuqi
from cuqi.distribution import Distribution
from copy import copy, deepcopy
import numpy as np


class RandomVariable:
    """ Random variable defined by a distribution with the option to apply algebraic operations on it.

    Random variables allow for the definition of Bayesian Problems in a natural way. In the context
    of code, the random variable can be viewed as a lazily evaluated variable/array. It records
    operations applied to it and acts as a function that, when called, evaluates the operations
    and returns the result.

    In CUQIpy, random variables can be in two forms: (1) a 'primal' random variable that is directly
    defined by a distribution, e.g. x ~ N(0, 1), or (2) a 'transformed' random variable that is defined by
    applying algebraic operations on one or more random variables, e.g. y = x + 1.

    This distinction is purely for the purpose of the implementation in CUQIpy, as mathematically both
    x ~ N(0, 1) and y = x + 1 ~ N(1, 1) are random variables. The distinction is useful for the
    code implementation. In the future some operations like the above may allow primal random variables
    that are transformed if the distribution can be analytically described.

    Parameters
    ----------
    distributions : Distribution or list of Distributions
        The distribution from which the random variable originates. If multiple distributions are
        provided, the random variable is defined by the passed abstract syntax `tree` representing the
        algebraic operations applied to one or more random variables.

    tree : Node, optional
        The tree, represented by the syntax tree nodes, that contain the algebraic operations applied to the random variable.
        Specifically, the root of the tree should be provided.

    name : str, optional
        Name of the random variable. If not provided, the name is extracted from either the distribution provided
        or from the variable name in the code. The name provided must match the parameter name of the distribution.

    Example
    -------

    Basic usage:

    .. code-block:: python

        from cuqi.distribution import Gaussian

        x = RandomVariable(Gaussian(0, 1))

    Defining Bayesian problem using random variables:

    .. code-block:: python

        from cuqi.testproblem import Deconvolution1D
        from cuqi.distribution import Gaussian, Gamma, GMRF
        from cuqi.experimental.algebra import RandomVariable
        from cuqi.problem import BayesianProblem

        import numpy as np
        A, y_obs, info = Deconvolution1D().get_components()

        # Bayesian problem
        d = RandomVariable(Gamma(1, 1e-4))
        s = RandomVariable(Gamma(1, 1e-4))
        x = RandomVariable(GMRF(np.zeros(A.domain_dim), d))
        y = RandomVariable(Gaussian(A @ x, 1/s))

        BP = BayesianProblem(y, x, s, d)
        BP.set_data(y=y_obs)
        BP.UQ()

    Defining random variable from multiple distributions:

    .. code-block:: python

        from cuqi.distribution import Gaussian, Gamma
        from cuqi.experimental.algebra import RandomVariable, VariableNode

        # Define the variables
        x = VariableNode('x')
        y = VariableNode('y')

        # Define the distributions (names must match variables)
        dist_x = Gaussian(0, 1, name='x')
        dist_y = Gamma(1, 1e-4, name='y')

        # Define the tree (this is the algebra that defines the random variable along with the distributions)
        tree = x + y

        # Define random variable from 2 distributions with relation x+y
        rv = RandomVariable([dist_x, dist_y], tree)

    """


    def __init__(self, distributions: Union['Distribution', List['Distribution']], tree: 'Node' = None, name: str = None):
        """ Create random variable from distribution """

        if isinstance(distributions, Distribution):
            distributions = [distributions]
        
        if not  isinstance(distributions, list) and not isinstance(distributions, _OrderedSet):
            raise ValueError("Expected a distribution or a list of distributions")

        # Convert single distribution(s) to internal datastructure _OrderedSet.
        # We use ordered set to ensure that the order of the distributions is preserved.
        # which in turn ensures that the parameter names are always in the same order.
        if not isinstance(distributions, _OrderedSet):
            distributions = _OrderedSet(distributions)

        # If tree is provided, check it is consistent with the given distributions
        if tree:
            tree_var_names = tree.get_variables()
            dist_par_names = {dist._name for dist in distributions}

            if len(tree_var_names) != len(distributions):
                raise ValueError(
                    f"There are {len(tree_var_names)} variables in the tree, but {len(distributions)} distributions are provided. "
                    "This may be due to passing multiple distributions with the same parameter name. "
                    f"The tree variables are {tree_var_names} and the distribution parameter names are {dist_par_names}."
                )
            
            if not all(var_name in dist_par_names for var_name in tree_var_names):
                raise ValueError(
                    f"Variable names in the tree {tree_var_names} do not match the parameter names in the distributions {dist_par_names}. "
                    "Ensure the name is inferred from the variable or explicitly provide it using name='var_name' in the distribution."
                )

        # Match random variable name with distribution parameter name (for single distribution)
        if len(distributions) == 1 and tree is None:
            dist = next(iter(distributions))
            dist_par_name = dist._name
            if dist_par_name is not None:
                if name is not None and dist_par_name != name:
                    raise ValueError(f"Parameter name '{dist_par_name}' of the distribution does not match the input name '{name}' for the random variable.")
                name = dist_par_name
    
        self._distributions = distributions
        """ The distribution from which the random variable originates. """
        
        self._tree = tree
        """ The tree representation of the random variable. """

        self._original_variable = None
        """ Stores the original variable if this is a conditioned copy"""

        self._name = name
        """ Name of the random variable. """


    def __call__(self, *args, **kwargs) -> Any:
        """ Evaluate random variable at a given parameter value. For example, for random variable `X`, `X(1)` gives `1` and `(X+1)(1)` gives `2` """

        if args and kwargs:
            raise ValueError("Cannot pass both positional and keyword arguments to RandomVariable")
        
        if args:
            kwargs = self._parse_args_add_to_kwargs(args, kwargs)

        # Check if kwargs match parameter names using a all compare
        if not all([name in kwargs for name in self.parameter_names]) or not all([name in self.parameter_names for name in kwargs]):
            raise ValueError(f"Expected arguments {self.parameter_names}, got arguments {kwargs}")

        return self.tree(**kwargs)

    def sample(self, N=1):
        """ Sample from the random variable. 

        Parameters
        ----------
        N : int, optional
            Number of samples to draw. Default is 1.
        """

        if self.is_cond:
            raise NotImplementedError(
                "Unable to directly sample from a random variable that has distributions with "
                "conditioning variables. This is not implemented."
            )

        if N == 1: return self(**{dist.name: dist.sample() for dist in self.distributions})
            
        samples = np.array([
            self(**{dist.name: dist.sample() for dist in self.distributions})
            for _ in range(N)
        ]).reshape(-1, N)  # Ensure correct shape (dim, N)

        return cuqi.samples.Samples(samples)
    
    @property
    def tree(self):
        if self._tree is None:
            if len(self._distributions) > 1:
                raise ValueError("Tree for multiple distributions can not be created automatically and need to be passed as an argument to the {} initializer.".format(type(self).__name__))
            self._tree = VariableNode(self.name)
        return self._tree

    @property
    def name(self):
        """ Name of the random variable. If not provided, the name is extracted from the variable name in the code. """
        if self._is_copy: # Extract the original variable name if this is a copy
            return self._original_variable.name
        if self._name is None: # If None extract the name from the stack
            self._name = cuqi.utilities._get_python_variable_name(self)
        if self._name is not None:
            self._inject_name_into_distribution(self._name)
        return self._name
    
    @name.setter
    def name(self, name):
        if self._is_copy:
            raise ValueError("This random variable is derived from the conditional random variable named "+self._original_variable.name+". The name of the derived random variable cannot be set, but follows the name of the original random variable.")
        self._name = name
        
    @property
    def distribution(self) -> cuqi.distribution.Distribution:
        """ Distribution from which the random variable originates. """
        if len(self._distributions) > 1:
            raise ValueError("Cannot get distribution from random variable defined by multiple distributions")
        self._inject_name_into_distribution()
        return next(iter(self._distributions))
    
    @property
    def distributions(self) -> set:
        """ Distributions from which the random variable originates. """
        self._inject_name_into_distribution()
        return self._distributions
        
    @property
    def parameter_names(self) -> str:
        """ Name of the parameter that the random variable can be evaluated at. """
        self._inject_name_into_distribution()
        return [distribution._name for distribution in self.distributions] # Consider renaming .name to .par_name for distributions

    @property
    def dim(self):
        if self.is_transformed:
            raise NotImplementedError("Dimension not implemented for transformed random variables")
        return self.distribution.dim

    @property
    def geometry(self):
        if self.is_transformed:
            raise NotImplementedError("Geometry not implemented for transformed random variables")
        return self.distribution.geometry

    @geometry.setter
    def geometry(self, geometry):
        if self.is_transformed:
            raise NotImplementedError("Geometry not implemented for transformed random variables")
        self.distribution.geometry = geometry

    @property
    def expression(self):
        """ Expression (formula) of the random variable. """
        return str(self.tree)

    @property
    def is_transformed(self):
        """ Returns True if the random variable is transformed. """
        return not isinstance(self.tree, VariableNode)

    @property
    def is_cond(self):
        """ Returns True if the random variable is a conditional random variable. """
        return any(dist.is_cond for dist in self.distributions)

    def condition(self, *args, **kwargs):
        """Condition the random variable on a given value. Only one of either positional or keyword arguments can be passed.
        
        Parameters
        ----------
        *args : Any
            Positional arguments to condition the random variable on. The order of the arguments must match the order of the parameter names.

        **kwargs : Any
            Keyword arguments to condition the random variable on. The keys must match the parameter names.
        
        """

        # Before conditioning, capture repr to ensure all variable names are injected
        self.__repr__()

        if args and kwargs:
            raise ValueError("Cannot pass both positional and keyword arguments to RandomVariable")
        
        if args:
            kwargs = self._parse_args_add_to_kwargs(args, kwargs)

        # Create a deep copy of the random variable to ensure the original tree is not modified
        new_variable = self._make_copy(deep=True)

        for kwargs_name in list(kwargs.keys()):
            value = kwargs.pop(kwargs_name)

            # Condition the tree turning the variable into a constant
            if kwargs_name in self.parameter_names:
                new_variable._tree = new_variable.tree.condition(**{kwargs_name: value})
            
            # Condition the random variable on both the distribution parameter name and distribution conditioning variables
            for dist in self.distributions:
                if kwargs_name == dist.name:
                    new_variable._remove_distribution(dist.name)
                elif kwargs_name in dist.get_conditioning_variables():
                    new_variable._replace_distribution(dist.name, dist(**{kwargs_name: value}))

        # Check if any kwargs are left unprocessed
        if kwargs:
            raise ValueError(f"Conditioning variables {list(kwargs.keys())} not found in the random variable {self}")

        return new_variable

    @property
    def _non_default_args(self) -> List[str]:
        """List of non-default arguments to distribution. This is used to return the correct
        arguments when evaluating the random variable.
        """
        return self.parameter_names

    def _replace_distribution(self, name, new_distribution):
        """ Replace distribution with a given name with a new distribution in the same position of the ordered set. """
        for dist in self.distributions:
            if dist._name == name:
                self._distributions.replace(dist, new_distribution)
                break

    def _remove_distribution(self, name):
        """ Remove distribution with a given name from the set of distributions. """
        for dist in self.distributions:
            if dist._name == name:
                self._distributions.remove(dist)
                break

    def _inject_name_into_distribution(self, name=None):
        if len(self._distributions) == 1:
            dist = next(iter(self._distributions))

            if dist._is_copy:
                dist = dist._original_density

            if dist._name is None:
                if name is None:
                    name = self.name
                dist.name = name # Inject using setter
    
    def _parse_args_add_to_kwargs(self, args, kwargs) -> dict:
        """ Parse args and add to kwargs if any. Arguments follow self.parameter_names order. """
        if len(args) != len(self.parameter_names):
            raise ValueError(f"Expected {len(self.parameter_names)} arguments, got {len(args)}. Parameters are: {self.parameter_names}")
        
        # Add args to kwargs
        for arg, name in zip(args, self.parameter_names):
            kwargs[name] = arg
            
        return kwargs

    def __repr__(self):
        # Create strings for parameter name ~ distribution pairs
        parameter_strings = [f"{name} ~ {distribution}" for name, distribution in zip(self.parameter_names, self.distributions)]
        # Join strings with newlines
        parameter_strings = "\n".join(parameter_strings)
        # Add initial newline and indentations
        parameter_strings = "\n".join(["\t"+line for line in parameter_strings.split("\n")])
        # Print parameter strings with newlines
        if self.is_transformed:
            title = f"Transformed Random Variable"
        else:
            title = f""
        if self.is_transformed:
            body = (
                f"\n"
                f"Expression: {self.tree}\n"
                f"Components: \n{parameter_strings}"
                )
        else:
            body = parameter_strings.replace("\t","")
        return title+body
   
    @property
    def _is_copy(self):
        """ Returns True if this is a copy of another random variable, e.g. by conditioning. """
        return hasattr(self, '_original_variable') and self._original_variable is not None

    def _make_copy(self, deep=False) -> 'RandomVariable':
        """ Returns a copy of the density keeping a pointer to the original. """
        if deep:
            new_variable = deepcopy(self)
            new_variable._original_variable = self
            return new_variable
        new_variable = copy(self)
        new_variable._distributions = copy(self.distributions)
        new_variable._tree = copy(self._tree)
        new_variable._original_variable = self
        return new_variable
    
    def _apply_operation(self, operation, other=None) -> 'RandomVariable':
        """
        Apply a specified operation to this RandomVariable.
        """
        if isinstance(other, cuqi.distribution.Distribution):
            raise ValueError("Cannot apply operation to distribution. Use .rv to create random variable first.")
        if other is None: # unary operation case
            return RandomVariable(self.distributions, operation(self.tree))
        elif isinstance(other, RandomVariable): # binary operation case with another random variable that has distributions
            return RandomVariable(self.distributions | other.distributions, operation(self.tree, other.tree))
        return RandomVariable(self.distributions, operation(self.tree, other)) # binary operation case with any other object (constant)

    def __add__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.add, other)

    def __radd__(self, other) -> 'RandomVariable':
        return self.__add__(other)

    def __sub__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.sub, other)

    def __rsub__(self, other) -> 'RandomVariable':
        return self._apply_operation(lambda x, y: operator.sub(y, x), other)

    def __mul__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.mul, other)

    def __rmul__(self, other) -> 'RandomVariable':
        return self.__mul__(other)

    def __truediv__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.truediv, other)

    def __rtruediv__(self, other) -> 'RandomVariable':
        return self._apply_operation(lambda x, y: operator.truediv(y, x), other)

    def __matmul__(self, other) -> 'RandomVariable':
        if isinstance(other, cuqi.model.Model) and not isinstance(other, cuqi.model.LinearModel):
            raise TypeError("Cannot apply matmul to non-linear models")
        return self._apply_operation(operator.matmul, other)

    def __rmatmul__(self, other) -> 'RandomVariable':
        if isinstance(other, cuqi.model.Model) and not isinstance(other, cuqi.model.LinearModel):
            raise TypeError("Cannot apply matmul to non-linear models")
        return self._apply_operation(lambda x, y: operator.matmul(y, x), other)

    def __neg__(self) -> 'RandomVariable':
        return self._apply_operation(operator.neg)

    def __abs__(self) -> 'RandomVariable':
        return self._apply_operation(abs)

    def __pow__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.pow, other)

    def __getitem__(self, other) -> 'RandomVariable':
        return self._apply_operation(operator.getitem, other)
